// Uebung4 Betriebssysteme 2
// Matthias Jauernig, 2006
// Server-Programm

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdbool.h>
#include <signal.h>
#include <sys/wait.h>

#define SOCKNAME "sockunix"
#define QUEUESIZE 10

#define Fflush() while(getchar() != '\n')

bool ende = false;

void sigintfunc(int signr){
	switch(signr){
	case SIGINT:
		printf("SIGINT empfangen...\n");
		ende = true;
	}
}

int main(void){
	int sockd, sockd2, l, sl, status, myPid;
	pid_t childPid;
	struct sockaddr_un uxadr; // Bereich fuer Adressinformation
	char eab[100]; // E/A-Bereich
	FILE *fp;
	char* fname = (char*)malloc(100*sizeof(char));

	// Adressinformationen bereitstellen
	uxadr.sun_family = AF_UNIX;
	strcpy(uxadr.sun_path, SOCKNAME);	// File fuer Datenaustausch eintragen

	// Socket einrichten
	if((sockd = socket(AF_UNIX, SOCK_STREAM, 0)) < 0){
		perror("socket");
		return 1;
	}

	// als Vorsichtsmaßnahme Socket-File loeschen - darf noch nicht existieren
	unlink(uxadr.sun_path);
	sl = sizeof(uxadr);
	if(bind(sockd, (struct sockaddr*)&uxadr, sl) < 0){ // Socketfile mit bind() anlegen
		perror("bind");
		return 2;
	}

	// Warteschlange fuer Clients anlegen
	if(listen(sockd, QUEUESIZE) < 0){
		perror("listen");
		return 3;
	}

	// Clients verarbeiten
	//signal(SIGINT, sigintfunc);	// SIGINT verarbeiten unter Linux
	sigset(SIGINT, sigintfunc);	// SIGINT verarbeiten unter Unix
	printf("Warte auf Client-Anforderungen...\n");
	do{
		// auf Client-Anforderungen warten
		if((sockd2 = accept(sockd, 0, 0)) < 0 && !ende){
			perror("accept");
			return 4;
		}

		if(ende){
			//close(sockd2);
			break;
		}

		// Child-Prozess erzeugen
		if((childPid = fork()) < 0){
			perror("fork");
			return 5;
		}
		switch(childPid){
			case 0: // im Child
				myPid = (int)getpid();
				printf("%d: Client hat sich verbunden...\n", myPid);
				if(close(sockd) < 0){
					perror("close(sockd)");
					return 6;
				}
				// Kommunikation ueber sockd2
				// 1.) Dateiname empfangen und Datei zum Schreiben oeffnen
				if((l = recv(sockd2, eab, 100, 0)) < 0){
					perror("recv");
				}
				strcpy(fname, eab);
				if((fp = fopen(eab, "w")) < 0){
					perror("fopen");
					return 8;
				}
				printf("%d: Schreibe Daten von Client in Datei '%s'\n", myPid, eab);
				
				// 2.) Daten empfangen und in Datei schreiben
				do{
					if((l = recv(sockd2, eab, 100, 0)) < 0){
						perror("recv");
					}
					if(l>0)
					{
						printf("%d: Daten von Client empfangen, %d Bytes: %s", myPid, l, eab);
						// Empfangene Daten in Datei schreiben
						if(fwrite(&eab, strlen(eab)+1, 1, fp) < 0){
							perror("fwrite");
							return 10;
						}
						// Byteanzahl als Quittung zurücksenden
						sprintf(eab, "%d", l);
						if((l=send(sockd2, eab, strlen(eab)+1, 0)) < 0){
							perror("send");
						}
					}
				}while(l != 0);
				printf("%d: Client hat sich verabschiedet...\n", myPid);
				fclose(fp);
				
				if(close(sockd2) < 0){
					perror("close(sockd2)");
					return 12;
				}
				
				return 0;
				break;
			default: // im Parent
				if(close(sockd2) < 0){
					perror("close(sockd2)");
					return 13;
				}
				break;
		}
	}while(!ende); // Ende ueber SIGINT

	// sockd schliessen
	if(close(sockd) < 0){
		perror("close(sockd)");
		return 15;
	}
	
	// auf Child-Ende warten
	while ((childPid=wait(&status)) > 0){
    		printf("Child %d beendet mit Status %x %x\n", (int)childPid, (status>>8)&0xFF, status&0xFF);
	}

	// Socket entfernen
	if(unlink(uxadr.sun_path) < 0){
		perror("unlink");
		return 14;
	}

	return 0;
}
